import os
import re
import unicodedata
import zlib
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Iterator

import fsspec


class AbstractTarget(ABC):
    @abstractmethod
    def get_mapper(self):
        pass

    @abstractmethod
    def exists(self, path: str) -> bool:
        """Check that the file exists."""
        pass

    @abstractmethod
    def rm(self, path: str) -> None:
        """Remove file."""
        pass

    @contextmanager
    def open(self, path: str, **kwargs):  # don't know how to type hint this
        """Open file with a context manager."""
        pass


def _hash_path(path: str) -> str:
    return str(hash(path))


@dataclass
class FSSpecTarget(AbstractTarget):
    """Representation of a storage target for Pangeo Forge.

    :param fs: The filesystem object we are writing to.
    :param root_path: The path under which the target data will be stored.
    """

    fs: fsspec.AbstractFileSystem
    root_path: str = ""

    def get_mapper(self) -> fsspec.mapping.FSMap:
        """Get a mutable mapping object suitable for storing Zarr data."""
        return self.fs.get_mapper(self.root_path)

    def _full_path(self, path: str):
        return os.path.join(self.root_path, path)

    def exists(self, path: str) -> bool:
        """Check that the file is in the cache."""
        return self.fs.exists(self._full_path(path))

    def rm(self, path: str) -> None:
        """Remove file from the cache."""
        self.fs.rm(self._full_path(path))

    @contextmanager
    def open(self, path: str, **kwargs) -> Iterator[None]:
        """Open file with a context manager."""
        with self.fs.open(self._full_path(path), **kwargs) as f:
            yield f

    def __post_init__(self):
        if not self.fs.isdir(self.root_path):
            self.fs.mkdir(self.root_path)


class FlatFSSpecTarget(FSSpecTarget):
    """A target that sanitizes all the path names so that everthing is stored
    in a single directory.

    Designed to be used as a cache for inputs.
    """

    def _full_path(self, path: str) -> str:
        # this is just in case _slugify(path) is non-unique
        prefix = hex(zlib.adler32(str(path).encode("utf8")))[2:10]
        slug = _slugify(path)
        new_path = "-".join([prefix, slug])
        return os.path.join(self.root_path, new_path)


class CacheFSSpecTarget(FlatFSSpecTarget):
    """Alias for FlatFSSpecTarget"""

    pass


def _slugify(value: str) -> str:
    # Adopted from
    # https://github.com/django/django/blob/master/django/utils/text.py
    # https://stackoverflow.com/questions/295135/turn-a-string-into-a-valid-filename
    value = str(value)
    value = unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    value = re.sub(r"[^.\w\s-]+", "_", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")


class UninitializedTarget(AbstractTarget):
    def get_mapper(self):
        raise UninitializedTargetError

    def exists(self, path: str) -> bool:
        raise UninitializedTargetError

    def rm(self, path: str) -> None:
        raise UninitializedTargetError

    def open(self, path: str, **kwargs):  # don't know how to type hint this
        raise UninitializedTargetError


class TargetError(Exception):
    """Base class for exceptions in this module."""

    pass


class UninitializedTargetError(TargetError):
    """Operation on an uninitialized Target."""

    pass



# import sphinx_pangeo_theme  # noqa
import sphinx_book_theme  # noqa

# -- Project information -----------------------------------------------------

project = "Pangeo Forge"
copyright = "2020, Pangeo Community"
author = "Pangeo Community"


# -- General configuration ---------------------------------------------------

extensions = [
    "myst_nb",
    "sphinx.ext.autodoc",
    # "numpydoc",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "**.ipynb_checkpoints"]
master_doc = "index"

# we always have to manually run the notebooks because they are slow / expensive
jupyter_execute_notebooks = "off"

# -- Options for HTML output -------------------------------------------------

# https://sphinx-book-theme.readthedocs.io/en/latest/configure.html
html_theme = "sphinx_book_theme"
html_theme_options = {
    "repository_url": "https://github.com/pangeo-forge/pangeo-forge",
    "repository_branch": "master",
    "path_to_docs": "docs",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_edit_page_button": True,
}
html_logo = "_static/pangeo-forge-logo-blue.png"
html_static_path = ["_static"]
myst_heading_anchors = 2
html_css_files = [
    "custom.css",
]



import base64
import http.server
import socketserver
import sys

port_str, ADDRESS = sys.argv[1:3]
PORT = int(port_str)
if len(sys.argv) > 3:
    username, password = sys.argv[3:5]
else:
    username, password = "", ""


class Handler(http.server.SimpleHTTPRequestHandler):
    def do_GET(self):
        if username:
            auth = self.headers.get("Authorization")
            if (
                auth is None
                or not auth.startswith("Basic")
                or auth[6:]
                != str(base64.b64encode((username + ":" + password).encode("utf-8")), "utf-8")
            ):
                self.send_response(401)
                self.send_header("WWW-Authenticate", "Basic")
                self.end_headers()
                return
        return http.server.SimpleHTTPRequestHandler.do_GET(self)


socketserver.TCPServer.allow_reuse_address = True
with socketserver.TCPServer((ADDRESS, PORT), Handler) as httpd:
    httpd.serve_forever()



import logging
import os
import socket
import subprocess
import time

import fsspec
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from dask.distributed import Client, LocalCluster

from pangeo_forge import recipe
from pangeo_forge.executors import (
    DaskPipelineExecutor,
    PrefectPipelineExecutor,
    PythonPipelineExecutor,
)
from pangeo_forge.patterns import VariableSequencePattern
from pangeo_forge.storage import CacheFSSpecTarget, FSSpecTarget, UninitializedTarget


def pytest_addoption(parser):
    parser.addoption(
        "--redirect-dask-worker-logs-to-stdout", action="store", default="NOTSET",
    )


def get_open_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = str(s.getsockname()[1])
    s.close()
    return port


@pytest.fixture(scope="session")
def daily_xarray_dataset():
    """Return a synthetic random xarray dataset."""
    np.random.seed(1)
    # TODO: change nt to 11 in order to catch the edge case where
    # items_per_input does not evenly divide the length of the sequence dimension
    nt, ny, nx = 10, 18, 36
    time = pd.date_range(start="2010-01-01", periods=nt, freq="D")
    lon = (np.arange(nx) + 0.5) * 360 / nx
    lon_attrs = {"units": "degrees_east", "long_name": "longitude"}
    lat = (np.arange(ny) + 0.5) * 180 / ny
    lat_attrs = {"units": "degrees_north", "long_name": "latitude"}
    foo = np.random.rand(nt, ny, nx)
    foo_attrs = {"long_name": "Fantastic Foo"}
    # make sure things work with heterogenous data types
    bar = np.random.randint(0, 10, size=(nt, ny, nx))
    bar_attrs = {"long_name": "Beautiful Bar"}
    dims = ("time", "lat", "lon")
    ds = xr.Dataset(
        {"bar": (dims, bar, bar_attrs), "foo": (dims, foo, foo_attrs)},
        coords={
            "time": ("time", time),
            "lat": ("lat", lat, lat_attrs),
            "lon": ("lon", lon, lon_attrs),
        },
        attrs={"conventions": "CF 1.6"},
    )
    return ds


def _split_up_files_by_day(ds, day_param):
    gb = ds.resample(time=day_param)
    _, datasets = zip(*gb)
    fnames = [f"{n:03d}.nc" for n in range(len(datasets))]
    return datasets, fnames


def _split_up_files_by_variable_and_day(ds, day_param):
    all_dsets = []
    all_fnames = []
    fnames_by_variable = {}
    for varname in ds.data_vars:
        var_dsets, fnames = _split_up_files_by_day(ds[[varname]], day_param)
        fnames = [f"{varname}_{fname}" for fname in fnames]
        all_dsets += var_dsets
        all_fnames += fnames
        fnames_by_variable[varname] = fnames
    return all_dsets, all_fnames, fnames_by_variable


@pytest.fixture(scope="session", params=["D", "2D"])
def netcdf_local_paths(daily_xarray_dataset, tmpdir_factory, request):
    """Return a list of paths pointing to netcdf files."""
    tmp_path = tmpdir_factory.mktemp("netcdf_data")
    # copy needed to avoid polluting metadata across multiple tests
    datasets, fnames = _split_up_files_by_day(daily_xarray_dataset.copy(), request.param)
    full_paths = [tmp_path.join(fname) for fname in fnames]
    xr.save_mfdataset(datasets, [str(path) for path in full_paths])
    items_per_file = {"D": 1, "2D": 2}[request.param]
    return full_paths, items_per_file


# TODO: this is quite repetetive of the fixture above. Replace with parametrization.
@pytest.fixture(scope="session", params=["D", "2D"])
def netcdf_local_paths_by_variable(daily_xarray_dataset, tmpdir_factory, request):
    """Return a list of paths pointing to netcdf files."""
    tmp_path = tmpdir_factory.mktemp("netcdf_data")
    datasets, fnames, fnames_by_variable = _split_up_files_by_variable_and_day(
        daily_xarray_dataset.copy(), request.param
    )
    full_paths = [tmp_path.join(fname) for fname in fnames]
    xr.save_mfdataset(datasets, [str(path) for path in full_paths])
    items_per_file = {"D": 1, "2D": 2}[request.param]
    path_format = str(tmp_path) + "/{variable}_{n:03d}.nc"
    return full_paths, items_per_file, fnames_by_variable, path_format


# TODO: refactor to allow netcdf_local_paths_by_variable to be passed without
# duplicating the whole test.
@pytest.fixture()
def netcdf_http_server(netcdf_local_paths, request):
    paths, items_per_file = netcdf_local_paths

    def make_netcdf_http_server(username="", password=""):
        first_path = paths[0]
        # assume that all files are in the same directory
        basedir = first_path.dirpath()
        fnames = [path.basename for path in paths]

        this_dir = os.path.dirname(os.path.abspath(__file__))
        port = get_open_port()
        command_list = [
            "python",
            os.path.join(this_dir, "http_auth_server.py"),
            port,
            "127.0.0.1",
            username,
            password,
        ]
        if username:
            command_list += [username, password]
        p = subprocess.Popen(command_list, cwd=basedir)
        url = f"http://127.0.0.1:{port}"
        time.sleep(1)  # let the server start up

        def teardown():
            p.kill()

        request.addfinalizer(teardown)
        return url, fnames, items_per_file

    return make_netcdf_http_server


@pytest.fixture()
def tmp_target(tmpdir_factory):
    fs = fsspec.get_filesystem_class("file")()
    path = str(tmpdir_factory.mktemp("target"))
    return FSSpecTarget(fs, path)


@pytest.fixture()
def tmp_cache(tmpdir_factory):
    path = str(tmpdir_factory.mktemp("cache"))
    fs = fsspec.get_filesystem_class("file")()
    cache = CacheFSSpecTarget(fs, path)
    return cache


@pytest.fixture()
def uninitialized_target():
    return UninitializedTarget()


@pytest.fixture
def netCDFtoZarr_sequential_recipe(daily_xarray_dataset, netcdf_local_paths, tmp_target, tmp_cache):
    paths, items_per_file = netcdf_local_paths
    kwargs = dict(
        input_urls=paths,
        sequence_dim="time",
        inputs_per_chunk=1,
        nitems_per_input=items_per_file,
        target=tmp_target,
        input_cache=tmp_cache,
    )
    return recipe.NetCDFtoZarrSequentialRecipe, kwargs, daily_xarray_dataset, tmp_target


@pytest.fixture
def netCDFtoZarr_sequential_multi_variable_recipe(
    daily_xarray_dataset, netcdf_local_paths_by_variable, tmp_target, tmp_cache
):
    paths, items_per_file, fnames_by_variable, path_format = netcdf_local_paths_by_variable
    nitems_per_input = items_per_file
    time_index = list(range(len(paths) // 2))
    pattern = VariableSequencePattern(
        path_format, keys={"variable": ["foo", "bar"], "n": time_index}
    )
    kwargs = dict(
        input_pattern=pattern,
        sequence_dim="time",
        inputs_per_chunk=1,
        nitems_per_input=nitems_per_input,
        target=tmp_target,
        input_cache=tmp_cache,
    )
    return recipe.NetCDFtoZarrMultiVarSequentialRecipe, kwargs, daily_xarray_dataset, tmp_target


@pytest.fixture(scope="session")
def dask_cluster(request):
    cluster = LocalCluster(n_workers=2, threads_per_worker=1, silence_logs=False)

    client = Client(cluster)

    # cluster setup

    def set_blosc_threads():
        from numcodecs import blosc

        blosc.use_threads = False

    log_level_name = request.config.getoption("--redirect-dask-worker-logs-to-stdout")
    level = logging.getLevelName(log_level_name)

    def redirect_logs():
        import logging

        logger = logging.getLogger("pangeo_forge")
        handler = logging.StreamHandler()
        handler.setLevel(level)
        logger.setLevel(level)
        logger.addHandler(handler)

    client.run(set_blosc_threads)
    client.run(redirect_logs)
    client.close()
    del client

    yield cluster

    cluster.close()


_executors = {
    "python": PythonPipelineExecutor,
    "dask": DaskPipelineExecutor,
    "prefect": PrefectPipelineExecutor,
    "prefect-dask": PrefectPipelineExecutor,
}


@pytest.fixture(params=["manual", "python", "dask", "prefect", "prefect-dask"])
def execute_recipe(request, dask_cluster):
    if request.param == "prefect-dask":
        # TODO: turn this off and fix this case!
        pytest.skip("Prefect with dask LocalCluster hangs inexplicably.")

    if request.param == "manual":

        def execute(r):
            if r.cache_inputs:
                for input_key in r.iter_inputs():
                    r.cache_input(input_key)
            r.prepare_target()
            for chunk_key in r.iter_chunks():
                r.store_chunk(chunk_key)
            r.finalize_target()

    else:
        ExecutorClass = _executors[request.param]

        def execute(rec):
            ex = ExecutorClass()
            pipeline = rec.to_pipelines()
            plan = ex.pipelines_to_plan(pipeline)

            if request.param == "dask":
                client = Client(dask_cluster)

            if request.param == "prefect-dask":
                from prefect.executors import DaskExecutor

                prefect_executor = DaskExecutor(address=dask_cluster.scheduler_address)
                plan.run(executor=prefect_executor)
            else:
                ex.execute_plan(plan)
            if request.param == "dask":
                client.close()
                del client

    return execute



"""
A Pangeo Forge Recipe
"""

import json
import logging
import os
import tempfile
from abc import ABC, abstractmethod
from contextlib import ExitStack, contextmanager
from dataclasses import dataclass, field
from typing import Callable, Dict, Hashable, Iterable, List, Optional

import dask
import fsspec
import numpy as np
import xarray as xr
import zarr
from rechunker.types import MultiStagePipeline, ParallelPipelines, Stage

from .patterns import ExplicitURLSequence, VariableSequencePattern
from .storage import AbstractTarget, UninitializedTarget, UninitializedTargetError
from .utils import (
    chunk_bounds_and_conflicts,
    chunked_iterable,
    fix_scalar_attr_encoding,
    lock_for_conflicts,
)

# use this filename to store global recipe metadata in the metadata_cache
# it will be written once (by prepare_target) and read many times (by store_chunk)
_GLOBAL_METADATA_KEY = "pangeo-forge-recipe-metadata.json"

logger = logging.getLogger(__name__)

# How to manually execute a recipe: ###
#
#   t = PangeoForgeTarget()
#   r = MyRecipe(target=t, **opts) # 1
#   # manual execution of recipe
#   for input_key in r.iter_inputs():
#       r.cache_input(input_key) # 4
#   r.prepare_target() # 3
#   for chunk_key in r.iter_chunks():
#       r.store_chunk(chunk_key) # 5
#   r.finalize_target() # 6


# 1) Initialize the Recipe object
# 2) Point the Recipe at its Target
# 3) Initialize the recipe.
#    Check if the target exists; if not, create it.
# 4) cache the inputs to proximate storage (OPTIONAL)
#    Some recipes won't need this (e.g. cloud to cloud)
#    If so, iter_inputs is just an empty iterator
# 5) Load each chunk from the inputs and store it in the target
#    Might be coming from the cache or might be read directly.
# 6)


class BaseRecipe(ABC):
    """Base recipe class from which all other Recipes inherit.
    """

    @property
    @abstractmethod
    def prepare_target(self) -> Callable[[], None]:
        """Prepare the recipe for execution by initializing the target.
        Attribute that returns a callable function.
        """
        pass

    @abstractmethod
    def iter_inputs(self) -> Iterable[Hashable]:
        """Iterate over all inputs."""
        pass

    @property
    @abstractmethod
    def cache_input(self) -> Callable[[Hashable], None]:
        """Copy an input from its source location to the cache.
        Attribute that returns a callable function.
        """
        pass

    @abstractmethod
    def iter_chunks(self) -> Iterable[Hashable]:
        """Iterate over all target chunks."""
        pass

    @property
    @abstractmethod
    def store_chunk(self) -> Callable[[Hashable], None]:
        """Store a chunk of data in the target.
        Attribute that returns a callable function.
        """
        pass

    @property
    @abstractmethod
    def finalize_target(self) -> Callable[[], None]:
        """Final step to finish the recipe after data has been written.
        Attribute that returns a callable function.
        """
        pass

    def to_pipelines(self) -> ParallelPipelines:
        """Translate recipe to pipeline for execution.
        """

        pipeline = []  # type: MultiStagePipeline
        if getattr(self, "cache_inputs", False):  # TODO: formalize this contract
            pipeline.append(Stage(self.cache_input, list(self.iter_inputs())))
        pipeline.append(Stage(self.prepare_target))
        pipeline.append(Stage(self.store_chunk, list(self.iter_chunks())))
        pipeline.append(Stage(self.finalize_target))
        pipelines = []  # type: ParallelPipelines
        pipelines.append(pipeline)
        return pipelines

    # https://stackoverflow.com/questions/59986413/achieving-multiple-inheritance-using-python-dataclasses
    def __post_init__(self):
        # just intercept the __post_init__ calls so they
        # aren't relayed to `object`
        pass


def _encode_key(key) -> str:
    return "-".join([str(k) for k in key])


def _input_metadata_fname(input_key):
    return "input-meta-" + _encode_key(input_key) + ".json"


def _chunk_metadata_fname(chunk_key) -> str:
    return "chunk-meta-" + _encode_key(chunk_key) + ".json"


def _copy_btw_filesystems(input_opener, output_opener, BLOCK_SIZE=10_000_000):
    with input_opener as source:
        with output_opener as target:
            while True:
                data = source.read(BLOCK_SIZE)
                if not data:
                    break
                target.write(data)


@contextmanager
def _maybe_open_or_copy_to_local(opener, copy_to_local, orig_name):
    _, suffix = os.path.splitext(orig_name)
    if copy_to_local:
        ntf = tempfile.NamedTemporaryFile(suffix=suffix)
        tmp_name = ntf.name
        logger.info(f"Copying {orig_name} to local file {tmp_name}")
        target_opener = open(tmp_name, mode="wb")
        _copy_btw_filesystems(opener, target_opener)
        yield tmp_name
        ntf.close()  # cleans up the temporary file
    else:
        with opener as fp:
            with fp as fp2:
                yield fp2


@contextmanager
def _fsspec_safe_open(fname, **kwargs):
    # workaround for inconsistent behavior of fsspec.open
    # https://github.com/intake/filesystem_spec/issues/579
    with fsspec.open(fname, **kwargs) as fp:
        with fp as fp2:
            yield fp2


# Notes about dataclasses:
# - https://www.python.org/dev/peps/pep-0557/#inheritance
# - https://stackoverflow.com/questions/51575931/class-inheritance-in-python-3-7-dataclasses
# The main awkward thing here is that, because we are using multiple inheritance
# with dataclasses, _ALL_ fields must have default values. This makes it impossible
# to have "required" keyword arguments--everything needs some default.
# That's whay we end up with `UninitializedTarget` and `_variable_sequence_pattern_default_factory`


@dataclass
class NetCDFtoZarrRecipe(BaseRecipe):
    """This class represents a dataset composed of many individual NetCDF files.
    This class uses Xarray to read and write data and writes its output to Zarr.
    The files are assumed to be arranged in a sequence along dimension ``sequence_dim``
    (commonly "time".)

    This class should not be used on its own but rather via one of its sub-classes.

    :param sequence_dim: The dimension name along which the inputs will be concatenated.
    :param inputs_per_chunk: The number of inputs to use in each chunk.
    :param nitems_per_input: The length of each input along the ``sequence_dim`` dimension.
      If ``None``, each input will be scanned and its metadata cached in order to
      determine the size of the target dataset.
    :param target: A location in which to put the dataset. Can also be assigned at run time.
    :param input_cache: A location in which to cache temporary data.
    :param metadata_cache: A location in which to cache metadata for inputs and chunks.
      Required if ``nitems_per_input=None``.
    :param cache_inputs: Whether to allow opening inputs directly which have not
      yet been cached. This could lead to very unstanble behavior if the inputs
      live behind a slow network connection.
    :param copy_input_to_local_file: Whether to copy the inputs to a temporary
      local file. In this case, a path (rather than file object) is passed to
      ``xr.open_dataset``. This is required for engines that can't open
      file-like objects (e.g. pynio).
    :param consolidate_zarr: Whether to consolidate the resulting Zarr dataset.
    :param xarray_open_kwargs: Extra options for opening the inputs with Xarray.
    :param xarray_concat_kwargs: Extra options to pass to Xarray when concatenating
      the inputs to form a chunk.
    :param delete_input_encoding: Whether to remove Xarray encoding from variables
      in the input dataset
    :param fsspec_open_kwargs: Extra options for opening the inputs with fsspec.
    :param process_input: Function to call on each opened input, with signature
      `(ds: xr.Dataset, filename: str) -> ds: xr.Dataset`.
    :param process_chunk: Function to call on each concatenated chunk, with signature
      `(ds: xr.Dataset) -> ds: xr.Dataset`.
    :param target_chunks: Desired chunk structure for the targret dataset.
    """

    sequence_dim: Optional[str] = None
    inputs_per_chunk: int = 1
    nitems_per_input: Optional[int] = 1
    target: AbstractTarget = field(default_factory=UninitializedTarget)
    input_cache: AbstractTarget = field(default_factory=UninitializedTarget)
    metadata_cache: AbstractTarget = field(default_factory=UninitializedTarget)
    cache_inputs: bool = True
    copy_input_to_local_file: bool = False
    consolidate_zarr: bool = True
    xarray_open_kwargs: dict = field(default_factory=dict)
    xarray_concat_kwargs: dict = field(default_factory=dict)
    delete_input_encoding: bool = True
    fsspec_open_kwargs: dict = field(default_factory=dict)
    process_input: Optional[Callable[[xr.Dataset, str], xr.Dataset]] = None
    process_chunk: Optional[Callable[[xr.Dataset], xr.Dataset]] = None
    target_chunks: Dict[str, int] = field(default_factory=dict)

    # private attributes, needed for type hints, not meant to be seen or accessed by user
    _sequence_dim_chunks: List[int] = field(default_factory=list, repr=False, init=False)
    _inputs: Dict[Hashable, str] = field(default_factory=dict, repr=False, init=False)

    def __post_init__(self):
        self._cache_metadata = self.nitems_per_input is None
        target_sequence_dim_chunks = self.target_chunks.get(self.sequence_dim)
        if (self.nitems_per_input is None) and (target_sequence_dim_chunks is None):
            raise ValueError(
                "Unable to determine target chunks. Please specify either "
                "`target_chunks` or `nitems_per_input`"
            )
        elif target_sequence_dim_chunks:
            self._sequence_dim_chunks = target_sequence_dim_chunks
        else:
            self._sequence_dim_chunks = self.nitems_per_input * self.inputs_per_chunk

        # TODO: more input validation
        # for example: check required args (e.g. sequence_dim)

    @property
    def prepare_target(self) -> Callable:
        def _prepare_target():

            try:
                ds = self.open_target()
                logger.info("Found an existing dataset in target")
                logger.debug(f"{ds}")

                if self.target_chunks:
                    # TODO: check that target_chunks id compatibile with the
                    # existing chunks
                    pass

            except (FileNotFoundError, IOError, zarr.errors.GroupNotFoundError):
                logger.info("Creating a new dataset in target")
                # need to rewrite this as an append loop
                for chunk_key in self._init_chunks:
                    with self.open_chunk(chunk_key) as ds:
                        # ds is already chunked

                        # https://github.com/pydata/xarray/blob/5287c7b2546fc8848f539bb5ee66bb8d91d8496f/xarray/core/variable.py#L1069
                        for v in ds.variables:
                            if self.target_chunks:
                                this_var = ds[v]
                                chunks = {
                                    this_var.get_axis_num(dim): chunk
                                    for dim, chunk in self.target_chunks.items()
                                    if dim in this_var.dims
                                }

                                chunks = tuple(
                                    chunks.get(n, s) for n, s in enumerate(this_var.shape)
                                )
                                encoding_chunks = chunks
                            else:
                                encoding_chunks = ds[v].shape
                            logger.debug(
                                f"Setting variable {v} encoding chunks to {encoding_chunks}"
                            )
                            ds[v].encoding["chunks"] = encoding_chunks

                        # load all variables that don't have the sequence dim in them
                        # these are usually coordinates.
                        # Variables that are loaded will be written even with compute=False
                        # TODO: make this behavior customizable
                        for v in ds.variables:
                            if self.sequence_dim not in ds[v].dims:
                                ds[v].load()

                        target_mapper = self.target.get_mapper()
                        logger.debug(f"Storing dataset:\n {ds}")
                        ds.to_zarr(target_mapper, mode="a", compute=False, safe_chunks=False)

            # Regardless of whether there is an existing dataset or we are creating a new one,
            # we need to expand the sequence_dim to hold the entire expected size of the data
            input_sequence_lens = self.sequence_lens()
            n_sequence = sum(input_sequence_lens)
            self.expand_target_dim(self.sequence_dim, n_sequence)

            if not self.nitems_per_input:
                # if nitems_per_input is not constant, we need to cache this info
                recipe_meta = {"input_sequence_lens": input_sequence_lens}
                meta_mapper = self.metadata_cache.get_mapper()
                # we are saving a dictionary with one key (input_sequence_lens)
                meta_mapper[_GLOBAL_METADATA_KEY] = json.dumps(recipe_meta).encode("utf-8")

        return _prepare_target

    @property
    def cache_input(self) -> Callable:
        def cache_func(input_key: Hashable) -> None:
            logger.info(f"Caching input {input_key}")
            fname = self._inputs[input_key]
            # TODO: check and see if the file already exists in the cache
            input_opener = _fsspec_safe_open(fname, mode="rb", **self.fsspec_open_kwargs)
            target_opener = self.input_cache.open(fname, mode="wb")
            _copy_btw_filesystems(input_opener, target_opener)
            if self._cache_metadata:
                self.cache_input_metadata(input_key)

        return cache_func

    @property
    def store_chunk(self) -> Callable:
        def _store_chunk(chunk_key):
            with self.open_chunk(chunk_key) as ds_chunk:
                # writing a region means that all the variables MUST have sequence_dim
                to_drop = [
                    v for v in ds_chunk.variables if self.sequence_dim not in ds_chunk[v].dims
                ]
                ds_chunk = ds_chunk.drop_vars(to_drop)

                target_mapper = self.target.get_mapper()
                write_region, conflicts = self.region_and_conflicts_for_chunk(chunk_key)

                zgroup = zarr.open_group(target_mapper)
                for vname, var_coded in ds_chunk.variables.items():
                    zarr_array = zgroup[vname]
                    # get encoding for variable from zarr attributes
                    # could this backfire some way?
                    var_coded.encoding.update(zarr_array.attrs)
                    # just delete all attributes from the var;
                    # they are not used anyway, and there can be conflicts
                    # related to xarray.coding.variables.safe_setitem
                    var_coded.attrs = {}
                    with dask.config.set(
                        scheduler="single-threaded"
                    ):  # make sure we don't use a scheduler
                        var = xr.backends.zarr.encode_zarr_variable(var_coded)
                        data = np.asarray(
                            var.data
                        )  # TODO: can we buffer large data rather than loading it all?
                    zarr_region = tuple(write_region.get(dim, slice(None)) for dim in var.dims)
                    lock_keys = [f"{vname}-{c}" for c in conflicts]
                    logger.debug(f"Acquiring locks {lock_keys}")
                    with lock_for_conflicts(lock_keys):
                        logger.info(
                            f"Storing variable {vname} chunk {chunk_key} "
                            f"to Zarr region {zarr_region}"
                        )
                        zarr_array[zarr_region] = data

        return _store_chunk

    @property
    def finalize_target(self) -> Callable:
        def _finalize_target():
            if self.consolidate_zarr:
                logger.info("Consolidating Zarr metadata")
                target_mapper = self.target.get_mapper()
                zarr.consolidate_metadata(target_mapper)

        return _finalize_target

    @contextmanager
    def input_opener(self, fname: str):
        try:
            logger.info(f"Opening '{fname}' from cache")
            opener = self.input_cache.open(fname, mode="rb")
            with _maybe_open_or_copy_to_local(opener, self.copy_input_to_local_file, fname) as fp:
                yield fp
        except (IOError, FileNotFoundError, UninitializedTargetError) as err:
            if self.cache_inputs:
                raise err(
                    f"You are trying to open input {fname}, but the file is "
                    "not cached yet. First call `cache_input` or set "
                    "`cache_inputs=False`."
                )
            logger.info(f"No cache found. Opening input `{fname}` directly.")
            opener = _fsspec_safe_open(fname, mode="rb", **self.fsspec_open_kwargs)
            with _maybe_open_or_copy_to_local(opener, self.copy_input_to_local_file, fname) as fp:
                yield fp

    @contextmanager
    def open_input(self, input_key: Hashable):
        fname = self._inputs[input_key]
        logger.info(f"Opening input with Xarray {input_key}: '{fname}'")
        with self.input_opener(fname) as f:
            logger.info(f"f is {f}")
            ds = xr.open_dataset(f, **self.xarray_open_kwargs)
            # Explicitly load into memory;
            # if we don't do this, we get a ValueError: seek of closed file.
            # But there will be some cases where we really don't want to load.
            # how to keep around the open file object?
            # ds = ds.load()

            ds = fix_scalar_attr_encoding(ds)

            if self.delete_input_encoding:
                for var in ds.variables:
                    ds[var].encoding = {}

            if self.process_input is not None:
                ds = self.process_input(ds, str(fname))

            logger.debug(f"{ds}")
            yield ds

    def cache_input_metadata(self, input_key: Hashable):
        logger.info(f"Caching metadata for input '{input_key}'")
        with self.open_input(input_key) as ds:
            metadata = ds.to_dict(data=False)
            mapper = self.metadata_cache.get_mapper()
            mapper[_input_metadata_fname(input_key)] = json.dumps(metadata).encode("utf-8")

    @contextmanager
    def open_chunk(self, chunk_key):
        logger.info(f"Concatenating inputs for chunk '{chunk_key}'")
        inputs = self.inputs_for_chunk(chunk_key)

        # need to open an unknown number of contexts at the same time
        with ExitStack() as stack:
            dsets = [stack.enter_context(self.open_input(i)) for i in inputs]
            # explicitly chunking prevents eager evaluation during concat
            dsets = [ds.chunk() for ds in dsets]
            if len(dsets) > 1:
                # During concat, attributes and encoding are taken from the first dataset
                # https://github.com/pydata/xarray/issues/1614
                ds = xr.concat(dsets, self.sequence_dim, **self.xarray_concat_kwargs)
            elif len(dsets) == 1:
                ds = dsets[0]
            else:  # pragma: no cover
                assert False, "Should never happen"

            if self.process_chunk is not None:
                ds = self.process_chunk(ds)

            logger.debug(f"{ds}")

            # TODO: maybe do some chunking here?
            yield ds

    def open_target(self):
        target_mapper = self.target.get_mapper()
        return xr.open_zarr(target_mapper)

    def expand_target_dim(self, dim, dimsize):
        target_mapper = self.target.get_mapper()
        zgroup = zarr.open_group(target_mapper)
        ds = self.open_target()
        sequence_axes = {v: ds[v].get_axis_num(dim) for v in ds.variables if dim in ds[v].dims}

        for v, axis in sequence_axes.items():
            arr = zgroup[v]
            shape = list(arr.shape)
            shape[axis] = dimsize
            logger.debug(f"resizing array {v} to shape {shape}")
            arr.resize(shape)

        # now explicity write the sequence coordinate to avoid missing data
        # when reopening
        if dim in zgroup:
            zgroup[dim][:] = 0

    def inputs_for_chunk(self, chunk_key):
        return self._chunks_inputs[chunk_key]

    def iter_inputs(self):
        for chunk_key in self.iter_chunks():
            for input in self.inputs_for_chunk(chunk_key):
                yield input

    def region_and_conflicts_for_chunk(self, chunk_key):
        # return a dict suitable to pass to xr.to_zarr(region=...)
        # specifies where in the overall array to put this chunk's data
        # also return the conflicts with other chunks

        input_keys = self.inputs_for_chunk(chunk_key)

        # TODO: refactor into a separate method
        if self.nitems_per_input:
            input_sequence_lens = (self.nitems_per_input,) * self._n_inputs_along_sequence
        else:
            input_sequence_lens = json.loads(
                self.metadata_cache.get_mapper()[_GLOBAL_METADATA_KEY]
            )["input_sequence_lens"]

        chunk_bounds, all_chunk_conflicts = chunk_bounds_and_conflicts(
            input_sequence_lens, self._sequence_dim_chunks
        )
        # for multi-variable recipes, there is something redunandt about this
        # logic that feels error prone
        start = chunk_bounds[self.input_position(input_keys[0])]
        stop = chunk_bounds[self.input_position(input_keys[-1]) + 1]

        this_chunk_conflicts = set()
        for k in input_keys:
            # for multi-variable recipes, the confilcts will usually be the same
            # for each variable. using a set avoids duplicate locks
            for input_conflict in all_chunk_conflicts[self.input_position(k)]:
                this_chunk_conflicts.add(input_conflict)
        region_slice = slice(start, stop)
        return {self.sequence_dim: region_slice}, this_chunk_conflicts

    def iter_chunks(self):
        for k in self._chunks_inputs:
            yield k

    def get_input_meta(self, *input_keys) -> Dict:
        meta_mapper = self.metadata_cache.get_mapper()
        # getitems should be async; much faster than serial calls
        all_meta_raw = meta_mapper.getitems([_input_metadata_fname(k) for k in input_keys])
        return {k: json.loads(raw_bytes) for k, raw_bytes in all_meta_raw.items()}

    def input_sequence_lens(self, *input_keys) -> List[int]:
        if self.nitems_per_input:
            return list((self.nitems_per_input,) * len(input_keys))  # struggling to please flake8
        # read per-input metadata; this is distinct from global metadata
        input_meta = self.get_input_meta(*input_keys)
        return [m["dims"][self.sequence_dim] for m in input_meta.values()]


@dataclass
class NetCDFtoZarrSequentialRecipe(NetCDFtoZarrRecipe):
    """There is only one sequence of input files. Each file can contain
    many variables.

    :param input_urls: The inputs used to generate the dataset.
    """

    input_urls: Iterable[str] = field(repr=False, default_factory=list)

    def __post_init__(self):
        super().__post_init__()
        input_pattern = ExplicitURLSequence(self.input_urls)
        self._inputs = {k: v for k, v in input_pattern}
        self._chunks_inputs = {
            (k,): v for k, v in enumerate(chunked_iterable(self._inputs, self.inputs_per_chunk))
        }
        self._n_inputs_along_sequence = len(self._inputs)

        # just the first chunk is needed to initialize the recipe
        self._init_chunks = [next(iter(self._chunks_inputs))]

    def input_position(self, input_key):
        return input_key[0]

    def chunk_position(self, chunk_key):
        return chunk_key[0]

    def sequence_lens(self):
        return self.input_sequence_lens(*self.iter_inputs())


def _variable_sequence_pattern_default_factory() -> VariableSequencePattern:
    return VariableSequencePattern(fmt_string="", keys={"variable": [], "_null": []})


@dataclass
class NetCDFtoZarrMultiVarSequentialRecipe(NetCDFtoZarrRecipe):
    """There are muliples sequences of input files (but all along the same dimension.)
    Different variables live in different files.

    :param input_pattern: An pattern used to generate the input file names.
    """

    input_pattern: VariableSequencePattern = field(
        default_factory=_variable_sequence_pattern_default_factory
    )

    def __post_init__(self):
        super().__post_init__()
        self._variables = self.input_pattern.keys["variable"]
        self._other_key = [k for k in self.input_pattern.keys if k != "variable"][0]
        self._inputs = {k: v for k, v in self.input_pattern}
        # input keys are tuples like
        #  ("temp", 0)
        #  ("temp", 1)
        #  ...
        #  ("salt", n)
        # chunk_keys will be similarly named, but the index will have a different meaning

        chunks_inputs = {}
        # generate this iteratively
        for vname in self._variables:
            # a hacky way to filter
            sequence_keys = [k[1] for k in self._inputs if k[0] == vname]
            for chunk_key, chunk_inputs in enumerate(
                chunked_iterable(sequence_keys, self.inputs_per_chunk)
            ):
                chunks_inputs[(vname, chunk_key)] = [
                    (vname, input_sequence_key) for input_sequence_key in chunk_inputs
                ]
        self._chunks_inputs = chunks_inputs
        self._n_inputs_along_sequence = len(self.input_pattern.keys[self._other_key])

        # should be the first chunk from each variable
        self._init_chunks = [chunk_key for chunk_key in self._chunks_inputs if chunk_key[1] == 0]

    def input_position(self, input_key):
        return self.input_pattern.keys[self._other_key].index(input_key[1])

    def chunk_position(self, chunk_key):
        return chunk_key[1]

    def sequence_lens(self):
        variables = self.input_pattern.keys["variable"]
        lens_by_variable = {}
        for v in variables:
            variable_input_keys = [k for k in self.iter_inputs() if k[0] == v]
            lens_by_variable[v] = self.input_sequence_lens(*variable_input_keys)
        if not all([lens_by_variable[v] == lens_by_variable[variables[0]]]):
            raise ValueError(f"Inconsistent sequence_len found for variables. f{lens_by_variable}")

        return lens_by_variable[variables[0]]



import pytest

from pangeo_forge.utils import chunk_bounds_and_conflicts, chunked_iterable


@pytest.mark.parametrize(
    "iterable, size, expected",
    [
        ([1, 2, 3], 1, [(1,), (2,), (3,)]),
        ([1, 2, 3], 2, [(1, 2), (3,)]),
        ([1, 2, 3], 3, [(1, 2, 3,)],),
        ([1, 2, 3], 4, [(1, 2, 3,)],),
    ],
)
def test_chunked_iterable(iterable, size, expected):
    actual = list(chunked_iterable(iterable, size))
    assert actual == expected


def test_chunk_conflicts():
    zchunks = 10
    assert chunk_bounds_and_conflicts([10, 10], zchunks) == ([0, 10, 20], [(), ()])
    assert chunk_bounds_and_conflicts([9, 10], zchunks) == ([0, 9, 19], [(0,), (0,)])
    assert chunk_bounds_and_conflicts([10, 9, 11, 10], zchunks) == (
        [0, 10, 19, 30, 40],
        [(), (1,), (1,), ()],
    )
    assert chunk_bounds_and_conflicts([9, 12, 5], zchunks) == ([0, 9, 21, 26], [(0,), (0, 2), (2,)])



from contextlib import nullcontext as does_not_raise

import pytest
import xarray as xr

# need to import this way (rather than use pytest.lazy_fixture) to make it work with dask
from pytest_lazyfixture import lazy_fixture

all_recipes = [
    lazy_fixture("netCDFtoZarr_sequential_recipe"),
    lazy_fixture("netCDFtoZarr_sequential_multi_variable_recipe"),
]


@pytest.mark.parametrize("recipe_fixture", all_recipes)
def test_recipe(recipe_fixture, execute_recipe):
    """The basic recipe test. Use this as a template for other tests."""

    RecipeClass, kwargs, ds_expected, target = recipe_fixture
    rec = RecipeClass(**kwargs)
    execute_recipe(rec)
    ds_actual = xr.open_zarr(target.get_mapper()).load()
    xr.testing.assert_identical(ds_actual, ds_expected)


@pytest.mark.parametrize("cache_inputs", [True, False])
@pytest.mark.parametrize("copy_input_to_local_file", [True, False])
def test_recipe_caching_copying(
    netCDFtoZarr_sequential_recipe, execute_recipe, cache_inputs, copy_input_to_local_file
):
    """The basic recipe test. Use this as a template for other tests."""

    RecipeClass, kwargs, ds_expected, target = netCDFtoZarr_sequential_recipe
    if not cache_inputs:
        kwargs.pop("input_cache")  # make sure recipe doesn't require input_cache
    rec = RecipeClass(
        **kwargs, cache_inputs=cache_inputs, copy_input_to_local_file=copy_input_to_local_file
    )
    execute_recipe(rec)
    ds_actual = xr.open_zarr(target.get_mapper()).load()
    xr.testing.assert_identical(ds_actual, ds_expected)


# function passed to preprocessing
def incr_date(ds, filename=""):
    # add one day
    t = [d + int(24 * 3600e9) for d in ds.time.values]
    ds = ds.assign_coords(time=t)
    return ds


@pytest.mark.parametrize(
    "process_input, process_chunk",
    [(None, None), (incr_date, None), (None, incr_date), (incr_date, incr_date)],
)
@pytest.mark.parametrize("recipe_fixture", all_recipes)
def test_process(recipe_fixture, execute_recipe, process_input, process_chunk):
    """Check that the process_chunk and process_input arguments work as expected."""

    RecipeClass, kwargs, ds_expected, target = recipe_fixture
    kwargs["process_input"] = process_input
    kwargs["process_chunk"] = process_chunk
    rec = RecipeClass(**kwargs)
    execute_recipe(rec)
    ds_actual = xr.open_zarr(target.get_mapper()).load()

    if process_input and process_chunk:
        assert not ds_actual.identical(ds_expected)
        ds_expected = incr_date(incr_date(ds_expected))
    elif process_input or process_chunk:
        assert not ds_actual.identical(ds_expected)
        ds_expected = incr_date(ds_expected)

    xr.testing.assert_identical(ds_actual, ds_expected)


@pytest.mark.parametrize("inputs_per_chunk", [1, 2])
@pytest.mark.parametrize(
    "target_chunks,specify_nitems_per_input,chunk_expectation",
    [
        ({}, True, does_not_raise()),
        ({"lon": 12}, True, does_not_raise()),
        ({"lon": 12, "time": 1}, True, does_not_raise()),
        ({"lon": 12, "time": 3}, True, does_not_raise()),
        ({"time": 100}, True, does_not_raise()),  # only one big chunk
        ({"lon": 12, "time": 1}, False, does_not_raise()),
        ({"lon": 12, "time": 3}, False, does_not_raise()),
        # can't determine target chunks for the next two because 'time' missing from target_chunks
        ({}, False, pytest.raises(ValueError)),
        ({"lon": 12}, False, pytest.raises(ValueError)),
    ],
)
@pytest.mark.parametrize("recipe_fixture", all_recipes)
def test_chunks(
    recipe_fixture,
    execute_recipe,
    inputs_per_chunk,
    target_chunks,
    chunk_expectation,
    specify_nitems_per_input,
):
    """Check that chunking of datasets works as expected."""

    RecipeClass, kwargs, ds_expected, target = recipe_fixture

    kwargs["target_chunks"] = target_chunks
    kwargs["inputs_per_chunk"] = inputs_per_chunk
    if specify_nitems_per_input:
        kwargs["nitems_per_input"] = kwargs["nitems_per_input"]  # it's already there in kwargs
        kwargs["metadata_cache"] = None
    else:
        # file will be scanned and metadata cached
        kwargs["nitems_per_input"] = None
        kwargs["metadata_cache"] = kwargs["input_cache"]

    with chunk_expectation as excinfo:
        rec = RecipeClass(**kwargs)
    if excinfo:
        # don't continue if we got an exception
        return

    execute_recipe(rec)

    # chunk validation
    ds_actual = xr.open_zarr(target.get_mapper(), consolidated=True)
    sequence_chunks = ds_actual.chunks["time"]
    seq_chunk_len = target_chunks.get("time", None) or (
        kwargs["nitems_per_input"] * inputs_per_chunk
    )
    # we expect all chunks but the last to have the expected size
    assert all([item == seq_chunk_len for item in sequence_chunks[:-1]])
    for other_dim, chunk_len in target_chunks.items():
        if other_dim == "time":
            continue
        assert all([item == chunk_len for item in ds_actual.chunks[other_dim][:-1]])

    ds_actual.load()
    print(ds_actual)
    xr.testing.assert_identical(ds_actual, ds_expected)



"""
Tests for our custom array writing locking stuff.
"""

import logging
import sys
from time import sleep

import dask
import numpy as np
import pytest
import zarr
from dask.distributed import Client

from pangeo_forge.utils import chunk_bounds_and_conflicts, lock_for_conflicts


@pytest.mark.parametrize("n_tasks, conflicts", [(2, (0,)), (2, (1, 2)), (4, (0,))])
def test_locks(n_tasks, conflicts, tmp_target, dask_cluster):

    this_client = Client(dask_cluster)

    @dask.delayed
    def do_stuff(n):
        logger = logging.getLogger("pangeo_forge")
        handler = logging.StreamHandler(stream=sys.stdout)
        handler.setLevel(logging.DEBUG)
        logger.addHandler(handler)
        with lock_for_conflicts(conflicts):
            # test for concurrency by writing an object and asserting there are
            # no other objects present
            mapper = tmp_target.get_mapper()
            key = f"foo_{n}"
            mapper[key] = b"bar"
            assert list(mapper) == [key]
            sleep(0.25)
            del mapper[key]
            assert list(mapper) == []

    dask.compute([do_stuff(n) for n in range(n_tasks)])

    this_client.close()
    del this_client


# chunking is always over first dimension
@pytest.mark.parametrize(
    "shape, zarr_chunks, write_chunks",
    [
        ((100,), (10,), (15,)),  # just a few conflicts
        ((100,), (100,), (10,)),  # lots of conflicts!
    ],
)
def test_locked_array_writing(shape, zarr_chunks, write_chunks, tmp_target, dask_cluster):

    sequence_lens = (shape[0] // write_chunks[0]) * [write_chunks[0]]
    remainder = shape[0] % write_chunks[0]
    if remainder > 0:
        sequence_lens.append(remainder)

    chunk_bounds, chunk_conflicts = chunk_bounds_and_conflicts(sequence_lens, zarr_chunks[0])

    data = np.random.rand(*shape)

    mapper = tmp_target.get_mapper()
    zarr_array = zarr.open(mapper, mode="w", shape=shape, chunks=zarr_chunks, dtype=data.dtype)

    @dask.delayed
    def write_block(n):
        write_region = slice(chunk_bounds[n], chunk_bounds[n + 1])
        with lock_for_conflicts(chunk_conflicts[n]):
            zarr_array[write_region] = data[write_region]

    this_client = Client(dask_cluster)
    dask.compute([write_block(n) for n in range(len(sequence_lens))])
    np.testing.assert_equal(data, zarr_array)

    this_client.close()
    del this_client



from rechunker.executors import DaskPipelineExecutor  # noqa: F401
from rechunker.executors import PrefectPipelineExecutor  # noqa: F401
from rechunker.executors import PythonPipelineExecutor  # noqa: F401



from pangeo_forge.pipeline import AbstractPipeline, XarrayPrefectPipelineMixin


class Pipeline(AbstractPipeline, XarrayPrefectPipelineMixin):
    def __init__(self, name="oisst-avhrr-v02r01"):
        self.name = name
        self.cache_location = f"gs://pangeo-scratch/rabernat/pangeo_smithy/{name}-cache/"
        self.concat_dim = "time"
        self.files_per_chunk = 5

    @property
    def sources(self):
        import pandas as pd

        keys = pd.date_range("1981-09-01", "1981-09-10", freq="D")
        source_url_pattern = (
            "https://www.ncei.noaa.gov/data/"
            "sea-surface-temperature-optimum-interpolation/v2.1/access/avhrr/"
            "{yyyymm}/oisst-avhrr-v02r01.{yyyymmdd}.nc"
        )
        source_urls = [
            source_url_pattern.format(yyyymm=key.strftime("%Y%m"), yyyymmdd=key.strftime("%Y%m%d"))
            for key in keys
        ]
        return source_urls

    @property
    def targets(self):
        return f"gs://pangeo-scratch/rabernat/pangeo_smithy/{self.name}-target/"



import os

import fsspec
import pandas as pd
import zarr
from prefect import Flow, task

from pangeo_forge.utils import chunked_iterable


def src_url_from_key(key):
    source_url_pattern = (
        "https://www.ncei.noaa.gov/data/sea-surface-temperature-optimum-interpolation"
        "/v2.1/access/avhrr/{yyyymm}/oisst-avhrr-v02r01.{yyyymmdd}.nc"
    )
    return source_url_pattern.format(yyyymm=key.strftime("%Y%m"), yyyymmdd=key.strftime("%Y%m%d"))


cache_location = "gs://pangeo-scratch/rabernat/pangeo_smithy/oisst-avhrr-v02r01-cache/"


@task
def download(key):
    source_url = src_url_from_key(key)
    target_url = os.path.join(cache_location, str(hash(key)))

    # there is probably a better way to do caching!
    try:
        fsspec.open(target_url).open()
        return target_url
    except FileNotFoundError:
        pass

    with fsspec.open(source_url, mode="rb") as source:
        with fsspec.open(target_url, mode="wb") as target:
            target.write(source.read())
    return target_url


@task
def combine_and_write(sources, target, append_dim, first=True):
    import xarray as xr

    # while debugging this, I had itermittent fsspec / hdf5 read errors related to
    # "trying to read from a closed file"
    # but they seem to have gone away for now
    double_open_files = [fsspec.open(url).open() for url in sources]
    ds = xr.open_mfdataset(double_open_files, combine="nested", concat_dim=concat_dim)
    # by definition, this should be a contiguous chunk
    ds = ds.chunk({append_dim: len(sources)})

    if first:
        kwargs = dict(mode="w")
    else:
        kwargs = dict(mode="a", append_dim=append_dim)

    mapper = fsspec.get_mapper(target)
    ds.to_zarr(mapper, **kwargs)


@task
def consolidate_metadata(target):
    mapper = fsspec.get_mapper(target)
    zarr.consolidate_metadata(mapper)


keys = pd.date_range("1981-09-01", "1981-09-10", freq="D")
target_path = "gs://pangeo-scratch/rabernat/pangeo_smithy/oisst-avhrr-v02r01-target/"
concat_dim = "time"
files_per_chunk = 5

with Flow("Pangeo-Forge") as flow:
    sources = [download(k) for k in keys]
    first = True
    write_tasks = []
    for source_group in chunked_iterable(sources, files_per_chunk):
        write_task = combine_and_write(source_group, target_path, concat_dim, first=first)
        write_tasks.append(write_task)
        first = False
    cm = consolidate_metadata(target_path)

# create dependencies in imperative mode
for n in range(1, len(write_tasks)):
    write_tasks[n].set_upstream(write_tasks[n - 1], flow=flow)
cm.set_upstream(write_tasks[-1], flow=flow)


# DO IT! #
flow.run()



from pkg_resources import DistributionNotFound, get_distribution

# from pangeo_forge.pipelines import AbstractPipeline

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:  # noqa: F401
    # package is not installed
    pass

del get_distribution, DistributionNotFound



"""
Filename patterns.
"""

import itertools
from abc import ABC
from dataclasses import dataclass
from typing import Dict, Generator, Hashable, Iterable, List, Tuple


class BasePattern(ABC):
    def __iter__(self) -> Generator[Tuple[Hashable, str], None, None]:
        pass


@dataclass
class URLPattern(BasePattern):
    """A generic pattern for representing a URL pattern

    :param fmt_string: A string that will be used to generate the URLs
    :param keys: Keys used to populate the template
    """

    fmt_string: str
    keys: Dict[str, Iterable]

    def __post_init__(self):
        pass
        # TODO: make a better check
        # this check does not work if you have to use the keys multiple times in the format string
        # fmt_string_keys = [
        #     t[1] for t in string.Formatter().parse(self.fmt_string) if t[1] is not None
        # ]
        # e.g. '{variable}/{variable[0]}.nc'
        # if not set(fmt_string_keys) == set(self.keys):
        #    raise KeyError("Specified keys don't match fmt_string")

    def __iter__(self):
        for keys in itertools.product(*self.keys.values()):
            format_kwargs = {k: v for k, v in zip(self.keys, keys)}
            yield keys, self.fmt_string.format(**format_kwargs)


@dataclass
class VariableSequencePattern(URLPattern):
    """A pattern for representing a dataset stored in a sequence (e.g. time)
    with different variables in different files.

    :param fmt_string: A string that will be used to generate the URLs
    :param keys: Keys used to populate the template. One key must be called ``variable``.
       The other can have any name. Only two keys are allowed.
    """

    def __post_init__(self):

        if not len(self.keys) == 2:
            raise ValueError("Exactly two keys are required for VariableSequencePattern")
        if "variable" not in self.keys:
            raise ValueError("keys must contain `variable` for VariableSequencePattern")
        self._sequence_key = [k for k in self.keys if k != "variable"][0]

        super().__post_init__()


@dataclass
class ExplicitURLSequence(BasePattern):
    """A pattern for representing a dataset stored in a sequence of URLs.

    :param urls: list of URLs.
    """

    urls: List[str]

    def __iter__(self):
        for n, item in enumerate(self.urls):
            yield (n,), item



import pytest

from pangeo_forge.patterns import ExplicitURLSequence, URLPattern, VariableSequencePattern


def test_url_pattern():
    pattern = URLPattern("{foo}:{bar}", {"foo": [1, 2], "bar": ["a", "b", "c"]})
    assert list(pattern) == [
        ((1, "a"), "1:a"),
        ((1, "b"), "1:b"),
        ((1, "c"), "1:c"),
        ((2, "a"), "2:a"),
        ((2, "b"), "2:b"),
        ((2, "c"), "2:c"),
    ]


def test_variable_sequence_pattern():
    pattern = VariableSequencePattern(
        "{variable}:{time}", {"variable": ["temp", "salt"], "time": [1, 2, 3]}
    )
    assert list(pattern) == [
        (("temp", 1), "temp:1"),
        (("temp", 2), "temp:2"),
        (("temp", 3), "temp:3"),
        (("salt", 1), "salt:1"),
        (("salt", 2), "salt:2"),
        (("salt", 3), "salt:3"),
    ]
    assert pattern._sequence_key == "time"

    with pytest.raises(ValueError, match=r".*`variable`.*"):
        _ = VariableSequencePattern("{foo}:{bar}", {"foo": ["temp", "salt"], "time": [1, 2, 3]})

    with pytest.raises(ValueError, match=r".*two keys.*"):
        _ = VariableSequencePattern(
            "{variable}:{time}:{foo}",
            {"variable": ["temp", "salt"], "time": [1, 2, 3], "foo": ["a"]},
        )


def test_explicit_url_sequence():
    seq = ExplicitURLSequence(["a", "b", "c"])
    assert list(seq) == [((0,), "a"), ((1,), "b"), ((2,), "c")]



import itertools
import logging
from contextlib import contextmanager
from typing import List, Sequence, Tuple

import numpy as np
from dask.distributed import Lock, get_client

logger = logging.getLogger(__name__)


# https://alexwlchan.net/2018/12/iterating-in-fixed-size-chunks/
def chunked_iterable(iterable, size):
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, size))
        if not chunk:
            break
        yield chunk


# only needed because of
# https://github.com/pydata/xarray/issues/4631
def fix_scalar_attr_encoding(ds):
    def _fixed_attrs(d):
        fixed = {}
        for k, v in d.items():
            if isinstance(v, np.ndarray) and len(v) == 1:
                fixed[k] = v[0]
        return fixed

    ds = ds.copy()
    ds.attrs.update(_fixed_attrs(ds.attrs))
    ds.encoding.update(_fixed_attrs(ds.encoding))
    for v in ds.variables:
        ds[v].attrs.update(_fixed_attrs(ds[v].attrs))
        ds[v].encoding.update(_fixed_attrs(ds[v].encoding))
    return ds


def chunk_bounds_and_conflicts(chunks: Sequence[int], zchunks: int) -> List[Tuple[int, ...]]:
    """
    Calculate the boundaries of contiguous put possibly uneven blocks over
    a regularly chunked array

    Parameters
    ----------
    chunks : A list of chunk lengths. Len of array is the sum of each length.
    zchunks : A constant on-disk chunk

    Returns
    -------
    chunk_bounds : the boundaries of the regions to write (1 longer than chunks)
    conflicts: a list of conflicts for each chunk, None for no conflicts
    """
    n_chunks = len(chunks)

    # coerce numpy array to list for mypy
    chunk_bounds = list([int(item) for item in np.hstack([0, np.cumsum(chunks)])])
    chunk_overlap = []
    for start, stop in zip(chunk_bounds[:-1], chunk_bounds[1:]):
        chunk_start = start // zchunks
        chunk_stop = (stop - 1) // zchunks
        chunk_overlap.append((chunk_start, chunk_stop))

    chunk_conflicts = []
    for n, chunk_pair in enumerate(chunk_overlap):
        conflicts = set()
        if n > 0:
            prev_pair = chunk_overlap[n - 1]
            if prev_pair[1] == chunk_pair[0]:
                conflicts.add(chunk_pair[0])
        if n < (n_chunks - 1):
            next_pair = chunk_overlap[n + 1]
            if next_pair[0] == chunk_pair[1]:
                conflicts.add(chunk_pair[1])
        chunk_conflicts.append(tuple(conflicts))

    return chunk_bounds, chunk_conflicts


@contextmanager
# TODO: use a recipe-specific base_name to handle multiple recipes potentially
# running at the same time
def lock_for_conflicts(conflicts, base_name="pangeo-forge"):

    try:
        global_client = get_client()
        is_distributed = True
    except ValueError:
        # Don't bother with locks if we are not in a distributed context
        # NOTE! This means we HAVE to use dask.distributed as our parallel execution enviroment
        # This should be compatible with Prefect.
        is_distributed = False
    if is_distributed:
        locks = [Lock(f"{base_name}-{c}", global_client) for c in conflicts]
        for lock in locks:
            logger.debug(f"Acquiring lock {lock.name}...")
            lock.acquire()
            logger.debug(f"Acquired lock {lock.name}")
    else:
        logger.debug(f"Asked to lock {conflicts} but no Dask client found.")
    try:
        yield
    finally:
        if is_distributed:
            for lock in locks:
                lock.release()
                logger.debug(f"Released lock {lock.name}")



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup

with open("requirements.txt") as f:
    install_requires = f.read().strip().split("\n")

with open("README.md") as f:
    long_description = f.read()


CLASSIFIERS = [
    "Development Status :: 1 - Planning",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: OS Independent",
    "Intended Audience :: Science/Research",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Topic :: Scientific/Engineering",
]

setup(
    name="pangeo-forge",
    description="Pipeline tools for building and publishing analysis ready datasets",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.7",
    maintainer="Joe Hamman",
    maintainer_email="jhamman@ucar.edu",
    classifiers=CLASSIFIERS,
    url="https://github.com/pangeo-forge/pangeo-forge",
    packages=find_packages(),
    package_dir={"pangeo_forge": "pangeo_forge"},
    include_package_data=True,
    install_requires=install_requires,
    license="Apache",
    zip_safe=False,
    keywords=["pangeo", "data"],
    entry_points=dict(console_scripts=["pangeo-forge = pangeo_forge.cli:main"]),
)



import fsspec
import pytest
import xarray as xr

from pangeo_forge.storage import UninitializedTargetError
from pangeo_forge.utils import fix_scalar_attr_encoding


def test_fixture_local_files(daily_xarray_dataset, netcdf_local_paths):
    paths, items_per_file = netcdf_local_paths
    paths = [str(path) for path in paths]
    ds = xr.open_mfdataset(paths, combine="nested", concat_dim="time")
    assert ds.identical(daily_xarray_dataset)


# TODO: this is quite repetetive of the test above. Replace with parametrization.
def test_fixture_local_files_by_variable(daily_xarray_dataset, netcdf_local_paths_by_variable):
    paths, items_per_file, fnames_by_variable, path_format = netcdf_local_paths_by_variable
    paths = [str(path) for path in paths]
    ds = xr.open_mfdataset(paths, combine="by_coords", concat_dim="time")
    assert ds.identical(daily_xarray_dataset)


def test_fixture_http_files(daily_xarray_dataset, netcdf_http_server):
    url, paths, items_per_file = netcdf_http_server()
    urls = ["/".join([url, str(path)]) for path in paths]
    open_files = [fsspec.open(url).open() for url in urls]
    ds = xr.open_mfdataset(open_files, combine="nested", concat_dim="time").load()
    ds = fix_scalar_attr_encoding(ds)
    assert ds.identical(daily_xarray_dataset)


def test_target(tmp_target):
    mapper = tmp_target.get_mapper()
    mapper["foo"] = b"bar"
    with open(tmp_target.root_path + "/foo") as f:
        res = f.read()
    assert res == "bar"
    with pytest.raises(FileNotFoundError):
        tmp_target.rm("baz")
    with pytest.raises(FileNotFoundError):
        with tmp_target.open("baz"):
            pass


def test_uninitialized_target(uninitialized_target):
    target = uninitialized_target
    with pytest.raises(UninitializedTargetError):
        target.get_mapper()
    with pytest.raises(UninitializedTargetError):
        target.exists("foo")
    with pytest.raises(UninitializedTargetError):
        target.rm("foo")
    with pytest.raises(UninitializedTargetError):
        with target.open("foo"):
            pass


def test_cache(tmp_cache):
    assert not tmp_cache.exists("foo")
    with tmp_cache.open("foo", mode="w") as f:
        f.write("bar")
    assert tmp_cache.exists("foo")
    with tmp_cache.open("foo", mode="r") as f:
        assert f.read() == "bar"
    tmp_cache.rm("foo")
    assert not tmp_cache.exists("foo")
